# RevenueCat V4 API Migration Guide
There were various updates to our API when we migrated the ObjC pieces to Swift. Most were unavoidable, 
but a number of updates make our Swift API more idomatic. We'll be updating this list as we continue to release betas.

To start us off, Our framework name changed from `Purchases` to `RevenueCat` 😻! We also updated all references of `Purchaser` to `Customer` to be more consistent across our platform. 

If you're using `Carthage`, make sure to use the new `RevenueCat.framework` or `RevenueCat.xcframework` instead of the old `Purchases`.

### Xcode version requirements and updated deployment targets
`purchases-ios` v4 requires using Xcode 13.2 or newer. 
It also updates the minimum deployment targets for iOS, macOS and tvOS. 

##### Minimum deployment targets
|  | v3 | v4 |
| :-: | :-: | :-: |
| iOS | 9.0 | 11.0 |
| tvOS | 9.0 | 11.0 |
| macOS | 10.12 | 10.13 |
| watchOS | 6.2 | 6.2 (unchanged) |

### Known Issues

#### ObjC + SPM
If you expose any Purchases objects from one target to another (see [example project](https://github.com/taquitos/SPMBug)
for what this could look like), that second target will not build due to a missing autogenerated header.
Currently there is a known bug in SPM whereby Xcode either doesn't pass the RevenueCat ObjC generated interface to SPM,
or SPM just doesn't integrate it. You can follow along: [SR-15154](https://bugs.swift.org/browse/SR-15154). 

##### Workaround: 
You must manually add the autogenerated file we committed to the repository
[RevenueCat-Swift.h](https://github.com/RevenueCat/purchases-ios/blob/main/IntegrationTests/CommonFiles/RevenueCat-Swift.h)
to your project, and `#import RevenueCat-Swift.h` in your bridging header. You can see how we do this in our
[SPMIntegration project](https://github.com/RevenueCat/purchases-ios/tree/main/IntegrationTests/SPMIntegration).

### Automatic Migration
When building your project using v4, Xcode should automatically provide one-click fixes methods and types that have been renamed. For the most part, the migration should be doable by just building and applying Xcode's automatic fix-its when they pop up.

If you see any issues or new APIs that fix-its didn't cover, we'd appreciate [bug reports](https://github.com/RevenueCat/purchases-ios/issues/new?assignees=&labels=bug&template=bug_report.md&title=)!

### New Types

To better support `StoreKit 2`, `RevenueCat v4` introduces several new types to encapsulate data from `StoreKit 1` and `StoreKit 2`:

- ``StoreProduct`` / `RCStoreProduct`: wraps a `StoreKit.SKProduct` or `StoreKit.Product`
- ``StoreTransaction`` / `RCStoreTransaction`: wraps a `StoreKit.SKPaymentTransaction` or `StoreKit.Transaction`
- ``StoreProductDiscount`` / `RCStoreProductDiscount`: wraps a `StoreKit.SKProductDiscount` or `StoreKit.Product.SubscriptionOffer`

These types replace native StoreKit types in all public API methods that used them.

### ObjC type changes

`@import Purchases` is now `@import RevenueCat`

| v3 | v4 |
| ------------ | ------------------------------------- | 
| RCPurchaserInfo | RCCustomerInfo |
| RCTransaction | RCStoreTransaction |
| RCPackage.product | RCPackage.storeProduct |
| (RCPurchasesErrorCode).RCOperationAlreadyInProgressError | RCOperationAlreadyInProgressForProductError |
| RCPurchasesErrorDomain | RCPurchasesErrorCodeDomain |
| RCBackendError | <i>REMOVED</i> |
| RCErrorUtils | <i>REMOVED</i> |
| RCBackendErrorDomain | <i>REMOVED</i> |
| RCFinishableKey | <i>REMOVED</i> |
| RCReceivePurchaserInfoBlock | <i>REMOVED</i> |
| RCReceiveIntroEligibilityBlock | <i>REMOVED</i> |
| RCReceiveOfferingsBlock | <i>REMOVED</i> |
| RCReceiveProductsBlock | <i>REMOVED</i> |
| RCPurchaseCompletedBlock | <i>REMOVED</i> |
| RCDeferredPromotionalPurchaseBlock | <i>REMOVED</i> |
| RCPaymentDiscountBlock | <i>REMOVED</i> |
| RCPaymentModeNone | <i>REMOVED</i> |

#### PurchasesDelegate
| v3 | v4 |
| ------------ | ------------------------------------- | 
| purchases:didReceiveUpdatedPurchaserInfo: | purchases:didReceiveUpdatedCustomerInfo: |

### ObjC API changes

| v3 | v4 |
| ------------ | ------------------------------------- | 
| purchaserInfoWithCompletion: | getCustomerInfoWithCompletion: |
| invalidatePurchaserInfoCache | invalidateCustomerInfoCache |
| Purchases -restoreTransactionsWithCompletion: | Purchases -restorePurchasesWithCompletion: |
| Purchases -offeringsWithCompletion: | Purchases -getOfferingsWithCompletion: |
| Purchases -productsWithIdentifiers:completion: | Purchases -getProductsWithIdentifiers:completion: |
| Purchases -paymentDiscountForProductDiscount:product:completion: | REMOVED - Check eligibility for a discount using `checkPromotionalOfferEligibility:`, then pass the discount directly to `purchasePackage:withDiscount:completion:` or `purchaseProduct:withDiscount:completion:` |
| Purchases -createAlias: | Purchases -logIn: |
| Purchases -identify: | Purchases -logIn: |
| Purchases -reset: | Purchases -logOut: |

### Swift type changes

`import Purchases` is now `import RevenueCat`

| v3 | v4 |
| ------------ | ------------------------------------- | 
| Purchases.Offering | ``Offering`` |
| Purchases.ErrorDomain | See error handling below |
| Purchases.ErrorCode.Code | See error handling below |
| Purchases.Package | ``Package`` |
| Purchases.PurchaserInfo | <strong>``CustomerInfo``</strong> |
| Purchases.EntitlementInfos | ``EntitlementInfos`` |
| Purchases.Transaction | ``StoreTransaction`` |
| Purchases.EntitlementInfo | ``EntitlementInfo`` |
| Purchases.PeriodType | ``PeriodType`` |
| Purchases.Store | ``Store`` |
| RCPurchaseOwnershipType | ``PurchaseOwnershipType`` |
| RCAttributionNetwork | ``AttributionNetwork`` |
| RCIntroEligibility | ``IntroEligibility`` |
| RCIntroEligibilityStatus | ``IntroEligibilityStatus`` |
| RCPaymentMode | ``StoreProductDiscount/PaymentMode-swift.enum`` |
| RCPaymentModeNone | <i>REMOVED</i> |
| Purchases.LogLevel | ``LogLevel`` |
| Purchases.Offerings | ``Offerings`` |
| Purchases.PackageType | ``PackageType`` |
| Package.product | ``Package/storeProduct`` |
| Package.product.price: NSDecimalNumber | ``StoreProduct/price``: Decimal |
| Package.localizedIntroductoryPriceString: String | ``Package/localizedIntroductoryPriceString``: String? |
| RCDeferredPromotionalPurchaseBlock | ``DeferredPromotionalPurchaseBlock`` |
| Purchases.PurchaseCompletedBlock | ``PurchaseCompletedBlock`` |
| Purchases.ReceivePurchaserInfoBlock | <i>REMOVED</i> |
| Purchases.ReceiveOfferingsBlock | <i>REMOVED</i> |
| Purchases.ReceiveProductsBlock | <i>REMOVED</i> |
| Purchases.PaymentDiscountBlock | <i>REMOVED</i> |
| Purchases.RevenueCatBackendErrorCode | <i>REMOVED</i> |
| Purchases.ErrorCode.operationAlreadyInProgressError | ``RevenueCat/ErrorCode/operationAlreadyInProgressForProductError`` |
| Purchases.ErrorUtils | <i>REMOVED</i> |
| ReadableErrorCodeKey | <i>REMOVED</i> |
| RCFinishableKey | <i>REMOVED</i> |

### Swift API changes

| v3 | v4 |
| ------------ | ------------------------------------- | 
| invalidatePurchaserInfoCache | ``Purchases/invalidateCustomerInfoCache()`` |
| logIn(_ appUserId:, _ completion:) | ``Purchases/logIn(_:completion:)`` |
| createAlias(_ alias:, _ completion:) | ``Purchases/logIn(_:completion:)`` |
| identify(_ appUserID:, _ completion:) | ``Purchases/logIn(_:completion:)`` |
| reset(completion:) | ``Purchases/logOut(completion:)`` |
| purchaserInfo(_ completion:) | ``Purchases/getCustomerInfo(completion:)`` |
| offerings(_ completion:) | ``Purchases/getOfferings(completion:)`` |
| products(_ productIdentifiers:, _ completion:) | ``Purchases/getProducts(_:completion:)`` |
| purchaseProduct(_ product:, _ completion:) | ``Purchases/purchase(product:completion:)`` |
| purchasePackage(_ package:, _ completion:) | ``Purchases/purchase(package:completion:)`` |
| restoreTransactions(_ completion:) | ``Purchases/restorePurchases(completion:)`` |
| syncPurchases(_ completion:) | ``Purchases/syncPurchases(completion:)`` |
| paymentDiscount(for:product:completion:) | REMOVED - Check eligibility for a discount using ``Purchases/checkPromotionalDiscountEligibility(forProductDiscount:product:)``, then pass the discount directly to ``Purchases/purchase(package:discount:)`` or ``Purchases/purchase(product:discount:)`` |
| purchaseProduct(_:discount:_) | ``Purchases/purchase(product:discount:completion:)`` |
| purchasePackage(_:discount:_) | ``Purchases/purchase(package:discount:completion:)`` |

#### PurchasesDelegate
| v3 | v4 |
| ------------ | ------------------------------------- | 
| purchases(_ purchases: Purchases, didReceiveUpdated purchaserInfo: PurchaserInfo) | ``PurchasesDelegate/purchases(_:receivedUpdated:)`` |

### Swift Error handling

Prior to the Swift migration, `Purchases` exposed errors as `NSError`s, so one could detect errors like this:
```swift
if error.domain == Purchases.ErrorDomain {
	switch Purchases.ErrorCode(_nsError: error).code {
		case .purchaseInvalidError: break
		default: break
	}
}
```
Starting from Version 4, this becomes much simpler:
```swift
if let error = error as? RevenueCat.ErrorCode {
	switch error {
		case .purchaseInvalidError: break
		default: break
	}
} else {
	// Error is a different type
}
```

### New APIs

- `showManageSuscriptions(completion:)`: Use this method to show the subscription management for the current user. Depending on where they made the purchase and their OS version, this might take them to the `managementURL`, or open the iOS Subscription Management page. 
- `beginRefundRequestForCurrentEntitlement`: Use this method to begin a refund request for the purchase that granted the current entitlement.
- `beginRefundRequest(forProduct:)`: Use this method to begin a refund request for a purchase, specifying the product identifier.
- `beginRefundRequest(forEntitlement:)`: Use this method to begin a refund request for a purchase, specifying the entitlement identifier.

## Reporting undocumented issues:

Feel free to file an issue! [New RevenueCat Issue](https://github.com/RevenueCat/purchases-ios/issues/new/).
